package org.hardknockjava.lesson20.example02.reader.broadway;


/******************************************************************
 * This is the same as BroadwayReaderJAXB, but instead of
 * reading the entire input file at once, it breaks off
 * <show> tags and parses them in turn.
 * On the output side, it writes the outer <broadway> tag,
 * then individual <show> tags, and finally the terminating
 * </broadway> tag.
 */

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Reader;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;

import org.hardknockjava.lesson20.example02.jaxb.broadway.Show;
import org.hardknockjava.lesson20.example02.jaxb.broadway.Show.Composers;
import org.hardknockjava.lesson20.example02.jaxb.broadway.Show.Lyricists;
import org.hardknockjava.lesson20.example02.jaxb.broadway.Show.Playwrights;

public class BroadwayReaderJAXBSingles {
    private static final String TAG_SHOW_START = "<show";
    private static final String TAG_SHOW_END = "</show>";
    
    // This statement gets the system line separator.
    private static final String LINE_SEPARATOR = System.getProperty("line.separator");

    private List<ShowPOJO> showList = new ArrayList<ShowPOJO>();
    private JAXBContext jaxbContext;
    private Unmarshaller jaxbUnmarshaller;
    private Marshaller jaxbMarshaller;

    public static void main(String[] args) throws Exception {
        BroadwayReaderJAXBSingles instance = new BroadwayReaderJAXBSingles();

        // Read the file and store it in instances of ShowInstance.
        instance.read("data/broadway.xml");

        // Dump the instances to the console.
        instance.dump();

        // Write the instances to the console.
        System.out.println();
        instance.write(System.out);
    }

    public BroadwayReaderJAXBSingles() throws Exception {
        // Note that while in the previous example the JAXB
        // context is based on Broadway, this one is
        // based on Show. That's because we're not going
        // to parse the entire contents of the <broadway>
        // tag. We're going to parse each of the <show>
        // tags in turn.
        jaxbContext = JAXBContext.newInstance(Show.class);
        jaxbUnmarshaller = jaxbContext.createUnmarshaller();
        jaxbMarshaller = jaxbContext.createMarshaller();
    }

    // Read the XML and store it.
    private void read(String fileName) throws Exception {
        File showFile = new File(fileName);

        // Establish a Reader to read the file,
        // and a StringBuilder to hold its contents.
        Reader rdr = new FileReader(showFile);
        StringBuilder buf = new StringBuilder();

        // Read into the buffer until we hit a <show> tag.
        // Then delete all characters that preceded it and
        // read until we have a </show> tag. Present the
        // enclosed characters to the parser.
        String tag;
        while ((tag = extractShowTag(rdr, buf)) != null) {
            addToProductionList(tag);
        }
    }

    // Get the next <show> tag from the input.
    private String extractShowTag(Reader rdr, StringBuilder buf) throws IOException {
        String returnString = null;
        int start = -1;

        // Read more into the buffer until a <show> tag appears
        // or until the input is exhausted.
        char[] input = new char[60];
        while ((start = buf.indexOf(TAG_SHOW_START)) < 0) {
            // We would normally read more than 60 characters
            // at a time, but we want to demonstrate the
            // technique of repeatedly reading until there's
            // a start tag--and if we read a lot of characters,
            // we'll probably get the end tag on the first try.
            int charsRead = rdr.read(input);
            
            // If the end of the file was reached before
            // finding a start tag, quit.
            if (charsRead < 0) {
                break;
            }
            
            // Add the characters we've read to the StringBuilder.
            buf.append(input, 0, charsRead);
        }

        // If no start tag is found, return a null.
        if (start < 1) {
            return null;
        }

        // Remove the characters up to the start tag.
        buf.delete(0, start);

        // Now search for the end tag.
        int end = -1;
        while ((end = buf.indexOf(TAG_SHOW_END)) < 0) {
            int charsRead = rdr.read(input);
            if (charsRead < 0) {
                break;
            }
            buf.append(input);
        }

        if (end < 0) {
            return null;
        }

        // Break off the complete <show> tag and return it.
        end += TAG_SHOW_END.length();

        returnString = buf.substring(0, end);
        
        // Remove the value we've returned from the StringBuilder.
        // What remains will be fed to the next call.
        buf.delete(0, end);
        
        return returnString;
    }

    // Marshal the <show> tag and add the result to the list.
    private void addToProductionList(String tag) throws Exception {
        InputStream xmlSource = new ByteArrayInputStream(tag.getBytes());
        Show show = (Show) jaxbUnmarshaller.unmarshal(xmlSource);

        String title = show.getTitle();
        String theatre = show.getTheatre();
        
        // Notice the peculiar method of getting a List
        // from JAXB. getPlaywrights() acquires the
        // <playwrights> tag; getPlaywright() gets
        // not a single <playwright>, but the List of
        // <playwright> tags.
        List<String> playwrights = show.getPlaywrights().getPlaywright();

        List<String> composers = new ArrayList<String>();
        if (show.getComposers() != null) {
            composers = show.getComposers().getComposer();
        }

        List<String> lyricists = new ArrayList<String>();
        if (show.getLyricists() != null) {
            lyricists = show.getLyricists().getLyricist();
        }

        Date opening = getNodeDateValue(show.getOpening());
        Date closing = getNodeDateValue(show.getClosing());
        Integer previews = show.getPreviews();
        Integer performances = show.getPerformances();
        ShowPOJO instance = new ShowPOJO(title, theatre, playwrights, composers, lyricists, opening, closing,
                previews, performances);
        showList.add(instance);
    }

    // This method is for elements that contain the string representation
    // of dates.
    private Date getNodeDateValue(String dateStr) {
        Date returnValue = null;

        if (dateStr != null) {
            SimpleDateFormat fmt = new SimpleDateFormat("MM/dd/yyyy");
            try {
                returnValue = fmt.parse(dateStr);
            } catch (Exception e) {
                returnValue = null;
            }
        }

        return returnValue;
    }

    // Dump the list of Shows to the console.
    private void dump() {
        for (ShowPOJO show : showList) {
            System.out.println(show.toString());
        }
    }

    // Return XML.
    private void write(OutputStream out) throws Exception {
        // We could use the DOM to generate the XML declaration,
        // but have chosen not to.
        out.write("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>".getBytes());
        
        // Add a line separator and the outer <broadway> tag.
        out.write(LINE_SEPARATOR.getBytes());
        out.write("<broadway>".getBytes());
        out.write(LINE_SEPARATOR.getBytes());

        // Build the <show> tags and append them to the StringBuilder.
        for (int index = showList.size() - 1; index >= 0; index--) {
            SimpleDateFormat fmt = new SimpleDateFormat("MM/dd/yyyy");

            ShowPOJO production = showList.get(index);
            Show showXml = new Show();

            if (production.getClosing() != null) {
                showXml.setClosing(fmt.format(production.getClosing()));
            }
            if (production.getOpening() != null) {
                showXml.setOpening(fmt.format(production.getOpening()));
            }

            if (!production.getComposerList().isEmpty()) {
                Composers composers = new Composers();
                composers.getComposer().addAll(production.getComposerList());
                showXml.setComposers(composers);
            }

            if (!production.getLyricistList().isEmpty()) {
                Lyricists lyricists = new Lyricists();
                lyricists.getLyricist().addAll(production.getLyricistList());
                showXml.setLyricists(lyricists);
            }

            Playwrights playwrights = new Playwrights();
            playwrights.getPlaywright().addAll(production.getPlaywrightList());
            showXml.setPlaywrights(playwrights);

            showXml.setPerformances(production.getPerformances());
            showXml.setPreviews(production.getPreviews());
            showXml.setTheatre(production.getTheatre());
            showXml.setTheatre(production.getTitle());

            // Marshal the JAXB Show instance to a ByteArrayOutputStream.
            OutputStream output = new ByteArrayOutputStream();
            
            // This statement suppresses the XML declaration, which
            // should appear only at the beginning of the output.
            jaxbMarshaller.setProperty(Marshaller.JAXB_FRAGMENT, true);
            
            // Create the XML in the output stream, which is in memory. 
            jaxbMarshaller.marshal(showXml, output);

            // Write the contents of the output stream to our
            // output device.
            out.write(output.toString().getBytes());
            out.write(LINE_SEPARATOR.getBytes());
        }

        // Append the final </broadway> tag.
        out.write("</broadway>".getBytes());
    }
}
